#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 6
void NES_mapper6::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,14,15);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
  else
  {
    parent_NES->ppu->PPU_VRAM_banks[0] = chr_ram + 0*0x400;
    parent_NES->ppu->PPU_VRAM_banks[1] = chr_ram + 1*0x400;
    parent_NES->ppu->PPU_VRAM_banks[2] = chr_ram + 2*0x400;
    parent_NES->ppu->PPU_VRAM_banks[3] = chr_ram + 3*0x400;
    parent_NES->ppu->PPU_VRAM_banks[4] = chr_ram + 4*0x400;
    parent_NES->ppu->PPU_VRAM_banks[5] = chr_ram + 5*0x400;
    parent_NES->ppu->PPU_VRAM_banks[6] = chr_ram + 6*0x400;
    parent_NES->ppu->PPU_VRAM_banks[7] = chr_ram + 7*0x400;
  }
}

void NES_mapper6::MemoryWriteLow(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x42FE:
      {
        if(data & 0x10)
        {
          set_mirroring(1,1,1,1);
        }
        else
        {
          set_mirroring(0,0,0,0);
        }
      }
      break;

    case 0x42FF:
      {
        if(data & 0x10)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;

    case 0x4501:
      {
        irq_enabled = 0;
      }
      break;

    case 0x4502:
      {
        irq_counter = (irq_counter & 0xFF00) | (uint32)data;
      }
      break;

    case 0x4503:
      {
        irq_counter = (irq_counter & 0x00FF) | ((uint32)data << 8);
        irq_enabled = 1;
      }
      break;
  }
}

void NES_mapper6::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = (data & 0x3C) >> 2;
  uint8 chr_bank = data & 0x03;

  set_CPU_bank4(prg_bank*2+0);
  set_CPU_bank5(prg_bank*2+1);

  parent_NES->ppu->PPU_VRAM_banks[0] = chr_ram + chr_bank*0x2000 + 0*0x400;
  parent_NES->ppu->PPU_VRAM_banks[1] = chr_ram + chr_bank*0x2000 + 1*0x400;
  parent_NES->ppu->PPU_VRAM_banks[2] = chr_ram + chr_bank*0x2000 + 2*0x400;
  parent_NES->ppu->PPU_VRAM_banks[3] = chr_ram + chr_bank*0x2000 + 3*0x400;
  parent_NES->ppu->PPU_VRAM_banks[4] = chr_ram + chr_bank*0x2000 + 4*0x400;
  parent_NES->ppu->PPU_VRAM_banks[5] = chr_ram + chr_bank*0x2000 + 5*0x400;
  parent_NES->ppu->PPU_VRAM_banks[6] = chr_ram + chr_bank*0x2000 + 6*0x400;
  parent_NES->ppu->PPU_VRAM_banks[7] = chr_ram + chr_bank*0x2000 + 7*0x400;
}

void NES_mapper6::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    irq_counter += 133;
    if(irq_counter >= 0xFFFF)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = 0;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 8
void NES_mapper8::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);

  // ignore frame-IRQ for Paris-Dakar Ralley Special
  if(parent_NES->sum_16k_rom() == 0x1de79b)
  {
    parent_NES->frame_irq_disenabled = 1;
  }
}

void NES_mapper8::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = (data & 0xF8) >> 3;
  uint8 chr_bank = data & 0x07;
  
  set_CPU_bank4(prg_bank*2+0);
  set_CPU_bank5(prg_bank*2+1);

  set_PPU_bank0(chr_bank*8+0);
  set_PPU_bank1(chr_bank*8+1);
  set_PPU_bank2(chr_bank*8+2);
  set_PPU_bank3(chr_bank*8+3);
  set_PPU_bank4(chr_bank*8+4);
  set_PPU_bank5(chr_bank*8+5);
  set_PPU_bank6(chr_bank*8+6);
  set_PPU_bank7(chr_bank*8+7);
}

/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 17
void NES_mapper17::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper17::MemoryWriteLow(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x42FE:
      {
        if((data & 0x10) == 0)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0x42FF:
      {
        if((data & 0x10) == 0)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
      }
      break;

    case 0x4501:
      {
        irq_enabled = 0;
      }
      break;

    case 0x4502:
      {
        irq_latch = (irq_latch & 0xFF00) | data;
      }
      break;

    case 0x4503:
      {
        irq_latch = (irq_latch & 0x00FF) | ((uint32)data << 8);
        irq_counter = irq_latch;
        irq_enabled = 1;
      }
      break;

    case 0x4504:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x4505:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x4506:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0x4507:
      {
        set_CPU_bank7(data);
      }
      break;

    case 0x4510:
      {
        set_PPU_bank0(data);
      }
      break;

    case 0x4511:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0x4512:
      {
        set_PPU_bank2(data);
      }
      break;

    case 0x4513:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0x4514:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0x4515:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0x4516:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0x4517:
      {
        set_PPU_bank7(data);
      }
      break;
  }
}

void NES_mapper17::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter >= 0xFFFF - 113)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = 0;
      irq_enabled = 0;
    }
    else
    {
      irq_counter += 113;
    }
  }
}
/////////////////////////////////////////////////////////////////////

#endif
