
/////////////////////////////////////////////////////////////////////
// Mapper 5
class NES_mapper5 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper5(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper5() {}

  void  Reset();

  uint8 MemoryReadLow(uint32 addr);
  void  MemoryWriteLow(uint32 addr, uint8 data);
  void  MemoryWriteSaveRAM(uint32 addr, uint8 data);
  void  MemoryWrite(uint32 addr, uint8 data);
  void  HSync(uint32 scanline);
  uint8 PPU_Latch_RenderScreen(uint8 mode, uint32 addr);

protected:
  void MMC5_set_CPU_bank(uint8 page, uint8 bank);
  void MMC5_set_WRAM_bank(uint8 page, uint8 bank);
  void sync_Chr_banks(uint8 page);

  uint32 wb[8];
  uint8 wram[8*0x2000];
  uint8 wram_size;

  uint8 chr_reg[8][2];

  uint8 irq_enabled;
  uint8 irq_status;
  uint32 irq_line;
  uint32 irq_counter;

  uint32 value0;
  uint32 value1;

  uint8 wram_protect0;
  uint8 wram_protect1;
  uint8 prg_size;
  uint8 chr_size;
  uint8 gfx_mode;
  uint8 split_control;
  uint8 split_bank;

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 10
class NES_mapper10 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper10(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper10() {}

  void  Reset();

  void  MemoryWrite(uint32 addr, uint8 data);

  void  PPU_Latch_FDFE(uint32 addr);

protected:
  uint8 regs[6];
  uint8 latch_0000;
  uint8 latch_1000;

  void set_VROM_0000();
  void set_VROM_1000();

  void SNSS_fixup(); // HACK HACK HACK HACK

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 47
class NES_mapper47 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper47(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper47() {}

  void  Reset();
  void  MemoryWriteSaveRAM(uint32 addr, uint8 data);
  void  MemoryWrite(uint32 addr, uint8 data);
  void  HSync(uint32 scanline);

protected:
  uint8  regs[8];
  uint32 rom_bank;
  uint32 prg0,prg1;
  uint32 chr01,chr23,chr4,chr5,chr6,chr7;
  uint32 chr_swap() { return regs[0] & 0x80; }
  uint32 prg_swap() { return regs[0] & 0x40; }
  uint8 irq_enabled; // IRQs enabled
  uint8 irq_counter; // IRQ scanline counter, decreasing
  uint8 irq_latch;   // IRQ scanline counter latch
  void MMC3_set_CPU_banks();
  void MMC3_set_PPU_banks();

private:
};

/////////////////////////////////////////////////////////////////////

