#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 42
void NES_mapper42::Reset()
{
  // set CPU bank pointers
  set_CPU_bank3(0);
  set_CPU_bank4(num_8k_ROM_banks-4);
  set_CPU_bank5(num_8k_ROM_banks-3);
  set_CPU_bank6(num_8k_ROM_banks-2);
  set_CPU_bank7(num_8k_ROM_banks-1);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper42::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE003)
  {
    case 0xE000:
      {
        set_CPU_bank3(data & 0x0F);
      }
      break;

    case 0xE001:
      {
        if(data & 0x08)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;

    case 0xE002:
      {
        if(data & 0x02)
        {
          irq_enabled = 1;
        }
        else
        {
          irq_enabled = 0;
          irq_counter = 0;
        }
      }
      break;
  }
}

void NES_mapper42::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter < 215)
    {
      irq_counter++;
    }
    if(irq_counter == 215)
    {
      parent_NES->cpu->DoIRQ();
      irq_enabled = 0;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 43
void NES_mapper43::Reset()
{
  set_CPU_banks(2,1,0,4,9);
  set_PPU_banks(0,1,2,3,4,5,6,7);

  //set_mirroring(NES_PPU::MIRROR_VERT);

  irq_enabled = 1;
  irq_counter = 0x2C;
  irq_latch = 0x2C;
}

uint8 NES_mapper43::MemoryReadLow(uint32 addr)
{
  if(0x5000 <= addr && addr < 0x6000)
  {
    return ROM_banks[0x2000*8 + 0x1000 + (addr - 0x5000)];
  }
  return (uint8)(addr >> 8);
}

void NES_mapper43::MemoryWriteLow(uint32 addr, uint8 data)
{
  if((addr & 0xF0FF) == 0x4022)
  {
    switch(data & 0x07)
    {
      case 0x0:
      case 0x2:
      case 0x3:
      case 0x4:
        {
          set_CPU_bank6(4);
        }
        break;

      case 0x1:
        {
          set_CPU_bank6(3);
        }
        break;

      case 0x5:
        {
          set_CPU_bank6(7);
        }
        break;

      case 0x6:
        {
          set_CPU_bank6(5);
        }
        break;

      case 0x7:
        {
          set_CPU_bank6(6);
        }
        break;
    }
  }
}

void NES_mapper43::MemoryWrite(uint32 addr, uint8 data)
{
  if(addr == 0x8122)
  {
    if((data & 0x03) == 0x03)
    {
      irq_enabled = 1;
      irq_counter = 0x23;
      irq_latch = 0x23;
    }
    else if((data & 0x03) == 0x00)
    {
      irq_enabled = 0;
    }
  }
}

void NES_mapper43::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter == 0)
    {
      irq_counter = irq_latch;
      parent_NES->cpu->DoIRQ();
    }
    else
    {
      irq_counter--;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 50
void NES_mapper50::Reset()
{
  set_CPU_banks(15,8,9,0,11);
  set_PPU_banks(0,1,2,3,4,5,6,7);

  irq_enabled = 0;
}

void NES_mapper50::MemoryWriteLow(uint32 addr, uint8 data)
{
  if((addr & 0xE060) == 0x4020)
  {
    if(addr & 0x0100)
    {
      irq_enabled = data & 0x01;
    }
    else
    {
      set_CPU_bank6((data&0x08)|((data&0x01)<<2)|((data&0x06)>>1));
    }
  }
}

void NES_mapper50::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(scanline == 21)
    {
      parent_NES->cpu->DoIRQ();
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 83
void NES_mapper83::Reset()
{
  regs[0] = regs[1] = regs[2] = 0;

  // set CPU bank pointers
  if(num_8k_ROM_banks >= 32)
  {
    set_CPU_banks(0,1,30,31);
    regs[1] = 0x30;
  }
  else
  {
    set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  }

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_enabled = 0;
  irq_counter = 0;
  chr_bank = 0;
}

uint8 NES_mapper83::MemoryReadLow(uint32 addr)
{
  if((addr & 0x5100) == 0x5100)
  {
    return regs[2];
  }
  else
  {
    return addr >> 8;
  }
}

void NES_mapper83::MemoryWriteLow(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x5101:
    case 0x5102:
    case 0x5103:
      {
        regs[2] = data;
      }
      break;
  }
}

void NES_mapper83::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
    case 0xB000:
    case 0xB0FF:
    case 0xB1FF:
      {
        regs[0] = data;
        chr_bank = (data & 0x30) << 4;
        set_CPU_bank4(data*2+0);
        set_CPU_bank5(data*2+1);
        set_CPU_bank6(((data&0x30)|0x0F)*2+0);
        set_CPU_bank7(((data&0x30)|0x0F)*2+1);
        
      }
      break;

    case 0x8100:
      {
        if(num_1k_VROM_banks <= 32*8)
        {
          regs[1] = data;
        }
        if((data & 0x03) == 0x00)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if((data & 0x03) == 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if((data & 0x03) == 0x02)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0x8200:
      {
        irq_counter = (irq_counter & 0xFF00) | (uint32)data;
      }
      break;

    case 0x8201:
      {
        irq_counter = (irq_counter & 0x00FF) | ((uint32)data << 8);
        irq_enabled = data;
      }
      break;

    case 0x8300:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x8301:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x8302:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0x8310:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank0(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank0((chr_bank^data)*2+0);
          set_PPU_bank1((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8311:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank1(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank2((chr_bank^data)*2+0);
          set_PPU_bank3((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8312:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank2(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank4((chr_bank^data)*2+0);
          set_PPU_bank5((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8313:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank3(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank6((chr_bank^data)*2+0);
          set_PPU_bank7((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8314:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank4(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank4((chr_bank^data)*2+0);
          set_PPU_bank5((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8315:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank5(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank6((chr_bank^data)*2+0);
          set_PPU_bank7((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8316:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank6(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank4((chr_bank^data)*2+0);
          set_PPU_bank5((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8317:
      {
        if((regs[1] & 0x30) == 0x30)
        {
          set_PPU_bank7(chr_bank^data);
        }
        else if((regs[1] & 0x30) == 0x10 || (regs[1] & 0x30) == 0x20)
        {
          set_PPU_bank6((chr_bank^data)*2+0);
          set_PPU_bank7((chr_bank^data)*2+1);
        }
      }
      break;

    case 0x8318:
      {
        set_CPU_bank4(((regs[0]&0x30)|data)*2+0);
        set_CPU_bank5(((regs[0]&0x30)|data)*2+1);
      }
      break;
  }
}

void NES_mapper83::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter <= 114)
    {
      parent_NES->cpu->DoIRQ();
      irq_enabled = 0;
    }
    else
    {
      irq_counter -= 114;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 90

void NES_mapper90::Reset()
{
  // set CPU bank pointers
  set_CPU_bank4(num_8k_ROM_banks-4);
  set_CPU_bank5(num_8k_ROM_banks-3);
  set_CPU_bank6(num_8k_ROM_banks-2);
  set_CPU_bank7(num_8k_ROM_banks-1);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);

  irq_counter = 0;
  irq_latch = 0;
  irq_enabled = 0;

  for(uint8 i = 0; i < 4; i++)
  {
    prg_reg[i] = num_8k_ROM_banks-4+i;
    nam_low_reg[i] = 0;
    nam_high_reg[i] = 0;
    chr_low_reg[i] = i;
    chr_high_reg[i] = 0;
    chr_low_reg[i+4] = i+4;
    chr_high_reg[i+4] = 0;
  }
}

uint8 NES_mapper90::MemoryReadLow(uint32 addr)
{
  if(addr == 0x5000)
  {
    return (uint8)(value1*value2 & 0x00FF);
  }
  else
  {
    return (uint8)(addr >> 8);
  }
}

void NES_mapper90::MemoryWriteLow(uint32 addr, uint8 data)
{
  if(addr == 0x5000)
  {
    value1 = data;
  }
  else if(addr == 0x5001)
  {
    value2 = data;
  }
}

void NES_mapper90::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
    case 0x8001:
    case 0x8002:
    case 0x8003:
      {
        prg_reg[addr & 0x03] = data;
        Sync_Prg_Banks();
      }
      break;

    case 0x9000:
    case 0x9001:
    case 0x9002:
    case 0x9003:
    case 0x9004:
    case 0x9005:
    case 0x9006:
    case 0x9007:
      {
        chr_low_reg[addr & 0x07] = data;
        Sync_Chr_Banks();
      }
      break;

    case 0xA000:
    case 0xA001:
    case 0xA002:
    case 0xA003:
    case 0xA004:
    case 0xA005:
    case 0xA006:
    case 0xA007:
      {
        chr_high_reg[addr & 0x07] = data;
        Sync_Chr_Banks();
      }
      break;

    case 0xB000:
    case 0xB001:
    case 0xB002:
    case 0xB003:
      {
        nam_low_reg[addr & 0x03] = data;
        Sync_Mirror();
      }
      break;

    case 0xB004:
    case 0xB005:
    case 0xB006:
    case 0xB007:
      {
        nam_high_reg[addr & 0x03] = data;
        Sync_Mirror();
      }
      break;

    case 0xC002:
      {
        irq_enabled = 0;
      }
      break;

    case 0xC003:
    case 0xC004:
      {
        if(irq_enabled == 0)
        {
          irq_enabled = 1;
          irq_counter = irq_latch;
        }
      }
      break;

    case 0xC005:
      {
        irq_counter = data;
        irq_latch = data;
      }
      break;

    case 0xD000:
      {
        prg_bank_6000 = data & 0x80;
        prg_bank_e000 = data & 0x04;
        prg_bank_size = data & 0x03;
        chr_bank_size = (data & 0x18) >> 3;
        mirror_mode = data & 0x20;
        Sync_Prg_Banks();
        Sync_Chr_Banks();
        Sync_Mirror();
      }
      break;

    case 0xD001:
      {
        mirror_type = data & 0x03;
        Sync_Mirror();
      }
      break;

    case 0xD003:
      {
        // bank page
      }
      break;
  }
}

void NES_mapper90::HSync(uint32 scanline)
{
  if((scanline >= 0) && (scanline <= 239))
  {
    if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
    {
      if(irq_counter == 0)
      {
        if(irq_enabled)
        {
          parent_NES->cpu->DoIRQ();
        }
        irq_latch = 0;
        irq_enabled = 0;
      }
      else
      {
        irq_counter--;
      }
    }
  }
}

void NES_mapper90::Sync_Mirror()
{
  uint8 i;
  uint32 nam_bank[4];

  for(i = 0; i < 4; i++)
  {
    nam_bank[i] = ((uint32)nam_high_reg[i] << 8) | (uint32)nam_low_reg[i];
  }

  if(mirror_mode)
  {
    for(i = 0; i < 4; i++)
    {
      if(!nam_high_reg[i] && (nam_low_reg[i] == i))
      {
        mirror_mode = 0;
      }
    }

    if(mirror_mode)
    {
      set_PPU_bank8(nam_bank[0]);
      set_PPU_bank9(nam_bank[1]);
      set_PPU_bank10(nam_bank[2]);
      set_PPU_bank11(nam_bank[3]);
    }
  }
  else
  {
    if(mirror_type == 0)
    {
      set_mirroring(NES_PPU::MIRROR_VERT);
    }
    else if(mirror_type == 1)
    {
      set_mirroring(NES_PPU::MIRROR_HORIZ);
    }
    else
    {
      set_mirroring(0,0,0,0);
    }
  }
}

void NES_mapper90::Sync_Chr_Banks()
{
  uint8 i;
  uint32 chr_bank[8];

  for(i = 0; i < 8; i++)
  {
    chr_bank[i] = ((uint32)chr_high_reg[i] << 8) | (uint32)chr_low_reg[i];
  }

  if(chr_bank_size == 0)
  {
    set_PPU_bank0(chr_bank[0]*8+0);
    set_PPU_bank1(chr_bank[0]*8+1);
    set_PPU_bank2(chr_bank[0]*8+2);
    set_PPU_bank3(chr_bank[0]*8+3);
    set_PPU_bank4(chr_bank[0]*8+4);
    set_PPU_bank5(chr_bank[0]*8+5);
    set_PPU_bank6(chr_bank[0]*8+6);
    set_PPU_bank7(chr_bank[0]*8+7);
  }
  else if(chr_bank_size == 1)
  {
    set_PPU_bank0(chr_bank[0]*4+0);
    set_PPU_bank1(chr_bank[0]*4+1);
    set_PPU_bank2(chr_bank[0]*4+2);
    set_PPU_bank3(chr_bank[0]*4+3);
    set_PPU_bank4(chr_bank[4]*4+0);
    set_PPU_bank5(chr_bank[4]*4+1);
    set_PPU_bank6(chr_bank[4]*4+2);
    set_PPU_bank7(chr_bank[4]*4+3);
  }
  else if(chr_bank_size == 2)
  {
    set_PPU_bank0(chr_bank[0]*2+0);
    set_PPU_bank1(chr_bank[0]*2+1);
    set_PPU_bank2(chr_bank[2]*2+0);
    set_PPU_bank3(chr_bank[2]*2+1);
    set_PPU_bank4(chr_bank[4]*2+0);
    set_PPU_bank5(chr_bank[4]*2+1);
    set_PPU_bank6(chr_bank[6]*2+0);
    set_PPU_bank7(chr_bank[6]*2+1);
  }
  else
  {
    set_PPU_bank0(chr_bank[0]);
    set_PPU_bank1(chr_bank[1]);
    set_PPU_bank2(chr_bank[2]);
    set_PPU_bank3(chr_bank[3]);
    set_PPU_bank4(chr_bank[4]);
    set_PPU_bank5(chr_bank[5]);
    set_PPU_bank6(chr_bank[6]);
    set_PPU_bank7(chr_bank[7]);
  }
}

void NES_mapper90::Sync_Prg_Banks()
{
  if(prg_bank_size == 0)
  {
    set_CPU_bank4(num_8k_ROM_banks-4);
    set_CPU_bank5(num_8k_ROM_banks-3);
    set_CPU_bank6(num_8k_ROM_banks-2);
    set_CPU_bank7(num_8k_ROM_banks-1);
  }
  else if(prg_bank_size == 1)
  {
    set_CPU_bank4(prg_reg[1]*2);
    set_CPU_bank5(prg_reg[1]*2+1);
    set_CPU_bank6(num_8k_ROM_banks-2);
    set_CPU_bank7(num_8k_ROM_banks-1);
  }
  else if(prg_bank_size == 2)
  {
    if(prg_bank_e000)
    {
      set_CPU_bank4(prg_reg[0]);
      set_CPU_bank5(prg_reg[1]);
      set_CPU_bank6(prg_reg[2]);
      set_CPU_bank7(prg_reg[3]);
    }
    else
    {
      if(prg_bank_6000)
      {
        set_CPU_bank3(prg_reg[3]);
      }
      set_CPU_bank4(prg_reg[0]);
      set_CPU_bank5(prg_reg[1]);
      set_CPU_bank6(prg_reg[2]);
      set_CPU_bank7(num_8k_ROM_banks-1);
    }
  }
  else
  {
    // 8k in reverse mode?
    set_CPU_bank4(prg_reg[3]);
    set_CPU_bank5(prg_reg[2]);
    set_CPU_bank6(prg_reg[1]);
    set_CPU_bank7(prg_reg[0]);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 91
void NES_mapper91::Reset()
{
  // set CPU bank pointers
  set_CPU_bank4(num_8k_ROM_banks-2);
  set_CPU_bank5(num_8k_ROM_banks-1);
  set_CPU_bank6(num_8k_ROM_banks-2);
  set_CPU_bank7(num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  set_mirroring(NES_PPU::MIRROR_VERT);
}

void NES_mapper91::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  switch(addr & 0xF00F)
  {
    case 0x6000:
      {
        set_PPU_bank0(data*2+0);
        set_PPU_bank1(data*2+1);
      }
      break;

    case 0x6001:
      {
        set_PPU_bank2(data*2+0);
        set_PPU_bank3(data*2+1);
      }
      break;

    case 0x6002:
      {
        set_PPU_bank4(data*2+0);
        set_PPU_bank5(data*2+1);
      }
      break;

    case 0x6003:
      {
        set_PPU_bank6(data*2+0);
        set_PPU_bank7(data*2+1);
      }
      break;

    case 0x7000:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x7001:
      {
        set_CPU_bank5(data);
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 112
void NES_mapper112::Reset()
{
  // clear registers FIRST!!!
  for(int i = 0; i < 8; i++) regs[i] = 0x00;

  // set CPU bank pointers
  prg0 = 0;
  prg1 = 1;
  MMC3_set_CPU_banks();

  // set VROM banks
  if(num_1k_VROM_banks)
  {
    chr01 = 0;
    chr23 = 2;
    chr4  = 4;
    chr5  = 5;
    chr6  = 6;
    chr7  = 7;
    MMC3_set_PPU_banks();
  }
  else
  {
    chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
  }

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper112::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE001)
  {
    case 0x8000:
      {
        regs[0] = data;
        MMC3_set_PPU_banks();
        MMC3_set_CPU_banks();
      }
      break;

    case 0xA000:
      {
        uint32 bank_num;
        regs[1] = data;
        bank_num = regs[1];
        switch(regs[0] & 0x07)
        {
          case 0x02:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr01 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x03:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr23 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x04:
            {
              if(num_1k_VROM_banks)
              {
                chr4 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x05:
            {
              if(num_1k_VROM_banks)
              {
                chr5 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x06:
            {
              if(num_1k_VROM_banks)
              {
                chr6 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x07:
            {
              if(num_1k_VROM_banks)
              {
                chr7 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x00:
            {
              prg0 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;

          case 0x01:
            {
              prg1 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;
        }
      }
      break;

    case 0x8001:
      {
        regs[2] = data;
        if(parent_NES->ROM->get_mirroring() != NES_PPU::MIRROR_FOUR_SCREEN)
        {
          if(data & 0x01)
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
        }
      }
      break;

    case 0xA001:
      {
        regs[3] = data;
      }
      break;

    case 0xC000:
      {
        regs[4] = data;
        irq_counter = regs[4];
      }
      break;

    case 0xC001:
      {
        regs[5] = data;
        irq_latch = regs[5];
      }
      break;

    case 0xE000:
      {
        regs[6] = data;
        irq_enabled = 0;
        if(data)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;

    case 0xE001:
      {
        regs[7] = data;
        irq_enabled = 1;
      }
      break;
  }
}

void NES_mapper112::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(irq_counter--))
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}

#define MAP112_ROM(ptr)  (((ptr)-parent_NES->ROM->get_ROM_banks())  >> 13)
#define MAP112_VROM(ptr) (((ptr)-parent_NES->ROM->get_VROM_banks()) >> 10)

void NES_mapper112::SNSS_fixup()
{
  NES_6502::Context context;
  parent_NES->cpu->GetContext(&context);

  prg0 = MAP112_ROM(context.mem_page[prg_swap() ? 6 : 4]);
  prg1 = MAP112_ROM(context.mem_page[5]);
  if(num_1k_VROM_banks)
  {
    if(chr_swap())
    {
      chr01 = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
      chr23 = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
      chr4  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
      chr5  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[1]);
      chr6  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
      chr7  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[3]);
    }
    else
    {
      chr01 = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
      chr23 = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
      chr4  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
      chr5  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[5]);
      chr6  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
      chr7  = MAP112_VROM(parent_NES->ppu->PPU_VRAM_banks[7]);
    }
  }
}

/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 113
void NES_mapper113::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper113::MemoryWriteLow(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x4100:
    case 0x4111:
    case 0x4120:
    case 0x4900:
      {
        uint8 prg_bank, chr_bank;

        prg_bank = data >> 3;
        if (num_8k_ROM_banks <= 8 && num_1k_VROM_banks == 8*16)
        {
          chr_bank = ((data >> 3) & 0x08) + (data & 0x07);
        }
        else
        {
          chr_bank = data & 0x07;
        }
        //uint8 prg_bank = (data & 0x03) | ((data & 0x80) >> 5);
        //uint8 chr_bank = (data & 0x70) >> 4;

        set_CPU_bank4(prg_bank*4);
        set_CPU_bank5(prg_bank*4+1);
        set_CPU_bank6(prg_bank*4+2);
        set_CPU_bank7(prg_bank*4+3);

        set_PPU_bank0(chr_bank*8);
        set_PPU_bank1(chr_bank*8+1);
        set_PPU_bank2(chr_bank*8+2);
        set_PPU_bank3(chr_bank*8+3);
        set_PPU_bank4(chr_bank*8+4);
        set_PPU_bank5(chr_bank*8+5);
        set_PPU_bank6(chr_bank*8+6);
        set_PPU_bank7(chr_bank*8+7);
      }
      break;
  }
}

void NES_mapper113::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8008:
    case 0x8009:
      {
        uint8 prg_bank, chr_bank;

        prg_bank = data >> 3;
        if (num_8k_ROM_banks <= 8 && num_1k_VROM_banks == 8*16)
        {
          chr_bank = ((data >> 3) & 0x08) + (data & 0x07);
        }
        else
        {
          chr_bank = data & 0x07;
        }
        //uint8 prg_bank = (data & 0x03) | ((data & 0x80) >> 5);
        //uint8 chr_bank = (data & 0x70) >> 4;

        set_CPU_bank4(prg_bank*4);
        set_CPU_bank5(prg_bank*4+1);
        set_CPU_bank6(prg_bank*4+2);
        set_CPU_bank7(prg_bank*4+3);

        set_PPU_bank0(chr_bank*8);
        set_PPU_bank1(chr_bank*8+1);
        set_PPU_bank2(chr_bank*8+2);
        set_PPU_bank3(chr_bank*8+3);
        set_PPU_bank4(chr_bank*8+4);
        set_PPU_bank5(chr_bank*8+5);
        set_PPU_bank6(chr_bank*8+6);
        set_PPU_bank7(chr_bank*8+7);
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 114
void NES_mapper114::Reset()
{
  // clear registers FIRST!!!
  for(int i = 0; i < 8; i++) regs[i] = 0x00;

  // set CPU bank pointers
  prg0 = 0;
  prg1 = 1;
  MMC3_set_CPU_banks();

  // set VROM banks
  if(num_1k_VROM_banks)
  {
    chr01 = 0;
    chr23 = 2;
    chr4  = 4;
    chr5  = 5;
    chr6  = 6;
    chr7  = 7;
    MMC3_set_PPU_banks();
  }
  else
  {
    chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
  }

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper114::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if((addr == 0x6000) && (data == 0x00))
  {
    for(int i = 0; i < 8; i++) regs[i] = 0x00;
    prg0 = 0;
    prg1 = 1;
    MMC3_set_CPU_banks();
    if(num_1k_VROM_banks)
    {
      chr01 = 0;
      chr23 = 2;
      chr4  = 4;
      chr5  = 5;
      chr6  = 6;
      chr7  = 7;
      MMC3_set_PPU_banks();
    }
    else
    {
      chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
    }
    irq_enabled = 0;
    irq_counter = 0;
    irq_latch = 0;
  }
}

void NES_mapper114::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE001)
  {
    case 0x8000:
      {
        regs[0] = data;
        MMC3_set_PPU_banks();
        MMC3_set_CPU_banks();
      }
      break;

    case 0x8001:
      {
        uint32 bank_num;
        regs[1] = data;
        bank_num = regs[1];
        switch(regs[0] & 0x07)
        {
          case 0x00:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr01 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x01:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr23 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x02:
            {
              if(num_1k_VROM_banks)
              {
                chr4 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x03:
            {
              if(num_1k_VROM_banks)
              {
                chr5 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x04:
            {
              if(num_1k_VROM_banks)
              {
                chr6 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x05:
            {
              if(num_1k_VROM_banks)
              {
                chr7 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x06:
            {
              prg0 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;

          case 0x07:
            {
              prg1 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;
        }
      }
      break;

    case 0xA000:
      {
        regs[2] = data;
        if(data & 0x40)
        {
          //LOG("MAP114 MIRRORING: 0x40 ???" << endl);
        }
        if(parent_NES->ROM->get_mirroring() != NES_PPU::MIRROR_FOUR_SCREEN)
        {
          if(data & 0x01)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
        }
      }
      break;

    case 0xA001:
      {
        regs[3] = data;
        if(data & 0x80)
        {
          // enable save RAM $6000-$7FFF
        }
        else
        {
          // disable save RAM $6000-$7FFF
        }
      }
      break;

    case 0xC000:
      regs[4] = data;
      irq_counter = regs[4];
      break;

    case 0xC001:
      regs[5] = data;
      irq_latch = regs[5];
      break;

    case 0xE000:
      regs[6] = data;
      irq_enabled = 0;
      break;

    case 0xE001:
      regs[7] = data;
      irq_enabled = 1;
      break;
  }
}

void NES_mapper114::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(irq_counter--))
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}

#define MAP114_ROM(ptr)  (((ptr)-parent_NES->ROM->get_ROM_banks())  >> 13)
#define MAP114_VROM(ptr) (((ptr)-parent_NES->ROM->get_VROM_banks()) >> 10)

void NES_mapper114::SNSS_fixup()
{
  NES_6502::Context context;
  parent_NES->cpu->GetContext(&context);

  prg0 = MAP114_ROM(context.mem_page[prg_swap() ? 6 : 4]);
  prg1 = MAP114_ROM(context.mem_page[5]);
  if(num_1k_VROM_banks)
  {
    if(chr_swap())
    {
      chr01 = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
      chr23 = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
      chr4  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
      chr5  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[1]);
      chr6  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
      chr7  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[3]);
    }
    else
    {
      chr01 = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
      chr23 = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
      chr4  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
      chr5  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[5]);
      chr6  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
      chr7  = MAP114_VROM(parent_NES->ppu->PPU_VRAM_banks[7]);
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 117
void NES_mapper117::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_line = 0;
  irq_enabled1 = 0;
  irq_enabled2 = 1;
}

void NES_mapper117::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x8001:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x8002:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0xA000:
      {
        set_PPU_bank0(data);
      }
      break;

    case 0xA001:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0xA002:
      {
        set_PPU_bank2(data);
      }
      break;

    case 0xA003:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0xA004:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0xA005:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0xA006:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0xA007:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0xA008:
    case 0xA009:
    case 0xA00a:
    case 0xA00b:
    case 0xA00c:
    case 0xA00d:
    case 0xA00e:
    case 0xA00f:
       break;

    case 0xC001:
    case 0xC002:
    case 0xC003:
      {
        irq_enabled1 = irq_line = data;
      }
      break;

    case 0xE000:
      {
        irq_enabled2 = data & 1;
      }
      break;
  }
}

void NES_mapper117::HSync(uint32 scanline)
{
  if(irq_enabled1 && irq_enabled2 && irq_line == scanline)
  {
      irq_enabled1 = 0;
      parent_NES->cpu->DoIRQ();
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 160
void NES_mapper160::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
  refresh_type = 0;
}

void NES_mapper160::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x8001:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x8002:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0x9000:
      {
        if(data == 0x2B)
        {
          refresh_type = 1; // title
        }
        else if(data == 0xA8)
        {
          refresh_type = 2; // pass word
        }
        else if(data == 0x1F)
        {
          refresh_type = 3; // game over
        }
        else if(data == 0x7C)
        {
          refresh_type = 4; // continue
        }
        else if(data == 0x18)
        {
          refresh_type = 5; // roulette
        }
        else if(data == 0x60)
        {
          refresh_type = 6; // congratulation
        }
        else
        {
          refresh_type = 0;
        }
        set_PPU_bank0(data);
      }
      break;

    case 0x9001:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0x9002:
      {
        if(refresh_type == 2 && data != 0xE8)
        {
          refresh_type = 0; // not pass word
        }
        set_PPU_bank2(data);
      }
      break;

    case 0x9003:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0x9004:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0x9005:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0x9006:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0x9007:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0xC000:
      {
        irq_counter = irq_latch;
        irq_enabled = irq_latch;
      }
      break;

    case 0xC001:
      {
        irq_latch = data;
      }
      break;

    case 0xC002:
      {
        irq_enabled = 0;
      }
      break;

    case 0xC003:
      {
        irq_counter = data;
      }
      break;
  }
}

void NES_mapper160::HSync(uint32 scanline)
{
  if(scanline == 0 || scanline == 239)
  {
    if(refresh_type == 1)
    {
      set_PPU_banks(0x58,0x59,0x5A,0x5B,0x58,0x59,0x5A,0x5B);
    }
    else if(refresh_type == 2)
    {
      set_PPU_banks(0x78,0x79,0x7A,0x7B,0x78,0x79,0x7A,0x7B);
    }
    else if(refresh_type == 3)
    {
      set_PPU_banks(0x7C,0x7D,0x7E,0x7F,0x7C,0x7D,0x7E,0x7F);
    }
    else if(refresh_type == 5)
    {
      set_PPU_banks(0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77);
    }
    else if(refresh_type == 6)
    {
      set_PPU_banks(0x5C,0x5D,0x5E,0x5F,0x7C,0x7D,0x7E,0x7F);
    }
  }
  if(scanline == 64)
  {
    if(refresh_type == 4)
    {
      // face of lamp's spirit
      if(parent_NES->ppu->PPU_VRAM_banks[8][32*10+16] == 0x0A)
      {
        set_PPU_bank0(0x68);
        set_PPU_bank1(0x69);
        set_PPU_bank2(0x6A);
        set_PPU_bank3(0x6B);
      }else{
        set_PPU_bank0(0x6C);
        set_PPU_bank1(0x6D);
        set_PPU_bank2(0x6E);
        set_PPU_bank3(0x6F);
      }
    }
  }
  if(scanline == 128)
  {
    if(refresh_type == 4)
    {
      set_PPU_bank0(0x68);
      set_PPU_bank1(0x69);
      set_PPU_bank2(0x6A);
      set_PPU_bank3(0x6B);
    }
    else if(refresh_type == 5)
    {
      set_PPU_banks(0x74,0x75,0x76,0x77,0x74,0x75,0x76,0x77);
    }
  }
  if(scanline == 160)
  {
    if(refresh_type == 6)
    {
      set_PPU_banks(0x60,0x61,0x5E,0x5F,0x7C,0x7D,0x7E,0x7F);
    }
  }
  if(irq_enabled)
  {
    if(irq_counter == 0xFF)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = 0;
      irq_enabled = 0;
    }
    else
    {
      irq_counter++;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 182
void NES_mapper182::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);

  regs[0] = 0;

  irq_enabled = 0;
  irq_counter = 0;
}

void NES_mapper182::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF003)
  {
    case 0x8001:
      {
        if(data & 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;

    case 0xA000:
      {
        regs[0] = data & 0x07;
      }
      break;

    case 0xC000:
      {
        switch(regs[0])
        {
          case 0x00:
            {
              set_PPU_bank0((data & 0xFE)+0);
              set_PPU_bank1((data & 0xFE)+1);
            }
            break;

          case 0x01:
            {
              set_PPU_bank5(data);
            }
            break;

          case 0x02:
            {
              set_PPU_bank2((data & 0xFE)+0);
              set_PPU_bank3((data & 0xFE)+1);
            }
            break;

          case 0x03:
            {
              set_PPU_bank7(data);
            }
            break;

          case 0x04:
            {
              set_CPU_bank4(data);
            }
            break;

          case 0x05:
            {
              set_CPU_bank5(data);
            }
            break;

          case 0x06:
            {
              set_PPU_bank4(data);
            }
            break;

          case 0x07:
            {
              set_PPU_bank6(data);
            }
            break;
        }
      }
      break;

    case 0xE003:
      {
        irq_counter = data;
        irq_enabled = data;
      }
      break;
  }
}

void NES_mapper182::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 240))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(--irq_counter))
        {
          irq_counter = 0;
          irq_enabled = 0;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////
// Mapper 187
void NES_mapper187::Reset()
{
  // clear registers FIRST!!!
  for(int i = 0; i < 8; i++) regs[i] = 0x00;

  // set CPU bank pointers
  prg0 = 0;
  prg1 = 1;
  MMC3_set_CPU_banks();

  // set VROM banks
  if(num_1k_VROM_banks)
  {
    chr01 = 0;
    chr23 = 2;
    chr4  = 4;
    chr5  = 5;
    chr6  = 6;
    chr7  = 7;
    MMC3_set_PPU_banks();
  }
  else
  {
    chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
  }

  ex_prg0 = 0;
  ex_prg1 = 1;
  ex_bank_enabled = ex_bank_mode = 0;

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper187::MemoryWriteLow(uint32 addr, uint8 data)
{
  //if(addr == 0x5000)
  {
    ex_bank_mode = data;

    if(data & 0x80)
    {
      if(data & 0x20)
      {
        set_CPU_bank4((data & 0x1E)*2+0);
        set_CPU_bank5((data & 0x1E)*2+1);
        set_CPU_bank6((data & 0x1E)*2+2);
        set_CPU_bank7((data & 0x1E)*2+3);
      }
      else
      {
        set_CPU_bank5((data & 0x1F)*2+0);
        set_CPU_bank6((data & 0x1F)*2+1);
      }
    }
    else
    {
      set_CPU_banks(ex_prg0,ex_prg1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
    }
  }
}

void NES_mapper187::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        ex_bank_enabled = 0;
        regs[0] = data;
        MMC3_set_PPU_banks();
        //MMC3_set_CPU_banks();
      }
      break;

    case 0x8001:
      {
        uint32 bank_num;
        regs[1] = data;
        bank_num = regs[1];

        if(ex_bank_enabled)
        {
          if(regs[0] == 0x2A)
          {
            set_CPU_bank5(0x0F);
          }
          if(regs[0] == 0x28)
          {
            set_CPU_bank6(0x17);
          }
        }
        else
        {
          switch(regs[0] & 0x07)
          {
            case 0x00:
              {
                if(num_1k_VROM_banks)
                {
                 bank_num &= 0xfe;
                  chr01 = bank_num;
                  MMC3_set_PPU_banks();
                }
              }
              break;

            case 0x01:
              {
                if(num_1k_VROM_banks)
                {
                  bank_num &= 0xfe;
                  chr23 = bank_num;
                  MMC3_set_PPU_banks();
                }
              }
              break;

            case 0x02:
              {
                if(num_1k_VROM_banks)
                {
                  chr4 = bank_num;
                  MMC3_set_PPU_banks();
                }
              }
              break;

            case 0x03:
              {
                if(num_1k_VROM_banks)
                {
                  chr5 = bank_num;
                  MMC3_set_PPU_banks();
                }
              }
              break;

            case 0x04:
              {
                if(num_1k_VROM_banks)
                {
                  chr6 = bank_num;
                  MMC3_set_PPU_banks();
                }
              }
              break;

            case 0x05:
              {
                if(num_1k_VROM_banks)
                {
                  chr7 = bank_num;
                  MMC3_set_PPU_banks();
                }
              }
              break;

            case 0x06:
              {
                if((ex_bank_mode & 0xA0) != 0xA0)
                {
                  prg0 = bank_num;
                  MMC3_set_CPU_banks();
                }
                ex_prg0 = data;
              }
              break;

            case 0x07:
              {
                if((ex_bank_mode & 0xA0) != 0xA0)
                {
                  prg1 = bank_num;
                  MMC3_set_CPU_banks();
                }
                ex_prg1 = data;
              }
              break;
          }
        }
        if((regs[0] & 0x07) == 0x06) ex_prg0 = data;
        if((regs[0] & 0x07) == 0x07) ex_prg1 = data;
      }
      break;

    case 0x8003:
      {
        ex_bank_enabled = 1;
        if(!(data & 0xF0))
        {
          set_CPU_bank6(num_8k_ROM_banks-2);
        }
      }
      break;

    case 0xA000:
      {
        regs[2] = data;

        if(parent_NES->ROM->get_mirroring() != NES_PPU::MIRROR_FOUR_SCREEN)
        {
          if(data & 0x01)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
        }
      }
      break;

    case 0xA001:
      {
        regs[3] = data;
      }
      break;

    case 0xC000:
      regs[4] = data;
      irq_counter = regs[4];
      break;

    case 0xC001:
      regs[5] = data;
      irq_latch = regs[5];
      break;

    case 0xE000:
    case 0xE002:
      regs[6] = data;
      irq_enabled = 0;
      break;

    case 0xE001:
    case 0xE003:
      regs[7] = data;
      irq_enabled = 1;
      break;
  }
}

void NES_mapper187::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(--irq_counter))
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}

void NES_mapper187::MMC3_set_CPU_banks()
{
  if(prg_swap())
  {
    set_CPU_banks(num_8k_ROM_banks-2,prg1,prg0,num_8k_ROM_banks-1);
  }
  else
  {
    set_CPU_banks(prg0,prg1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  }
}

void NES_mapper187::MMC3_set_PPU_banks()
{
  if(num_1k_VROM_banks)
  {
    if(chr_swap())
    {
      set_PPU_banks(chr4,chr5,chr6,chr7,chr01,chr01+1,chr23,chr23+1);
    }
    else
    {
      set_PPU_banks(chr01,chr01+1,chr23,chr23+1,chr4,chr5,chr6,chr7);
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 189
void NES_mapper189::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_counter = 0;
  irq_latch = 0;
  irq_enabled = 0;
}

void NES_mapper189::MemoryWriteLow(uint32 addr, uint8 data)
{
  if(addr >= 0x4100 && addr <= 0x41FF)
  {
    data = (data & 0x30) >> 4;
    set_CPU_banks(data*4,data*4+1,data*4+2,data*4+3);
  }
}

void NES_mapper189::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        regs[0] = data;
      }
      break;

    case 0x8001:
      {
        switch(regs[0])
        {
          case 0x40:
            {
              set_PPU_bank0(data+0);
              set_PPU_bank1(data+1);
            }
            break;

          case 0x41:
            {
              set_PPU_bank2(data+0);
              set_PPU_bank3(data+1);
            }
            break;

          case 0x42:
            {
              set_PPU_bank4(data);
            }
            break;

          case 0x43:
            {
              set_PPU_bank5(data);
            }
            break;

          case 0x44:
            {
              set_PPU_bank6(data);
            }
            break;

          case 0x45:
            {
              set_PPU_bank7(data);
            }
            break;

          case 0x46:
            {
              set_CPU_bank6(data);
            }
            break;

          case 0x47:
            {
              set_CPU_bank5(data);
            }
            break;
        }
      }
      break;

    case 0xC000:
      {
        irq_counter = data;
      }
      break;

    case 0xC001:
      {
        irq_latch = data;
      }
      break;

    case 0xE000:
      {
        irq_enabled = 0;
      }
      break;

    case 0xE001:
      {
        irq_enabled = 1;
      }
      break;
  }
}

void NES_mapper189::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(--irq_counter))
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 243
void NES_mapper243::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  regs[0] = regs[1] = regs[2] = regs[3] = 0;
}

void NES_mapper243::MemoryWriteLow(uint32 addr, uint8 data)
{
  if(addr == 0x4100)
  {
    regs[0] = data;
  }
  else if(addr == 0x4101)
  {
    switch(regs[0] & 0x07)
    {
      case 0x02:
        {
          regs[1] = data & 0x01;
        }
        break;

      case 0x00:
      case 0x04:
      case 0x07:
        {
          regs[2] = (data & 0x01) << 1;
        }
        break;

      case 0x05:
        {
          set_CPU_bank4(data*4+0);
          set_CPU_bank5(data*4+1);
          set_CPU_bank6(data*4+2);
          set_CPU_bank7(data*4+3);
        }
        break;

      case 0x06:
        {
          regs[3] = (data & 0x03) << 2;
        }
        break;
    }
    if (num_1k_VROM_banks <= 64)
    {
      uint8 chr_bank = (regs[2] + regs[3]) >> 1;
      set_PPU_bank0(chr_bank*8+0);
      set_PPU_bank1(chr_bank*8+1);
      set_PPU_bank2(chr_bank*8+2);
      set_PPU_bank3(chr_bank*8+3);
      set_PPU_bank4(chr_bank*8+4);
      set_PPU_bank5(chr_bank*8+5);
      set_PPU_bank6(chr_bank*8+6);
      set_PPU_bank7(chr_bank*8+7);
    }
    else
    {
      uint8 chr_bank = regs[1] + regs[2] + regs[3];
      set_PPU_bank0(chr_bank*8+0);
      set_PPU_bank1(chr_bank*8+1);
      set_PPU_bank2(chr_bank*8+2);
      set_PPU_bank3(chr_bank*8+3);
      set_PPU_bank4(chr_bank*8+4);
      set_PPU_bank5(chr_bank*8+5);
      set_PPU_bank6(chr_bank*8+6);
      set_PPU_bank7(chr_bank*8+7);
    }
  }
}
/////////////////////////////////////////////////////////////////////

#endif
